@@ -17,35 +17,257 @@ function set_data( counter, key, uri ) {
   d.setAttribute('ADM_URI', encodeURI(uri));
 }
 
+function find_parent( o, criteria ) {
+  while( o ) {
+    var found=true;
+    for( var i in criteria ) {
+      if( criteria[i] instanceof RegExp ) {
+	if( !criteria[i].test(o[i]) ) {
+	  found=false;
+	  break;
+	}
+      } else if( criteria[i] instanceof String ||
+		 typeof(criteria[i])=='string' ||
+		 typeof(criteria[i])=='number' ) {
+	if( o[i]!=criteria[i] ) {
+	  found=false;
+	  break;
+	}
+      } else if( criteria[i] instanceof Function ) {
+	if( !criteria[i](o[i]) ) {
+	  found=false;
+	  break;
+	}
+      }
+    }
+    if( found ) return o;
+    o=o.parentNode;
+  }
+  return null;
+}
+
+function add_outer_shortcuts() {
+  var el=document.getElementsByTagName('body')[0];
+  var func;
+
+  func = function(e) {
+    e = e || window.event;
+			
+    //Find Which key is pressed
+    if (typeof(e.keyCode)=='number') code = e.keyCode;
+    else if (typeof(e.which)=='number') code = e.which;
+    var src=e.target!=null?e.target:e.srcElement;
+
+    var propagate=true;
+    if(src!=null && e.ctrlKey) {
+      var o;
+      var c;
+      c=find_parent(src, {id: /^form(\d+)$/});
+      if( c!=null ) {
+	c=c.id.match(/^form(\d+)$/)[1];
+      }
+      if( !e.shiftKey && !e.altKey && !e.metaKey ) {
+	switch(code) {
+	case 83:		// ctrl+s: save
+	  if( c!=null ) {
+	    o=$('save'+c);
+	    if( o.style.visibility!='hidden' ) xupdate(c, o, src);
+	  }
+	  propagate=false;
+	  break;
+	  
+	case 190:		// ctrl+.: insert new action below
+	  o=find_parent(src, {tagName: "TABLE", className: "inner_tdc"});
+	  if( o ) xinsert( o.parentNode, 1 );
+	  propagate=false;
+	  break;
+	  
+	case 188:		// ctrl+,: insert new action above
+	  o=find_parent(src, {tagName: "TABLE", className: "inner_tdc"});
+	  if( o ) xinsert( o.parentNode, -1 );
+	  propagate=false;
+	  break;
+	}
+      } else if( e.shiftKey && !e.altKey && !e.metaKey ) {
+	switch(code) {
+	case 67:		// ctrl+shift+c: close
+	  if( c!=null ) xclose(c, $('a'+c));
+	  propagate=false;
+	  break;
+
+	case 85:		// ctrl+shift+u: reload
+	  if( c!=null ) xreload(c, $('reload'+c), src);
+	  propagate=false;
+	  break;
+
+	case 190:		// ctrl+shift+.: insert new block below
+	  o=find_parent(src, {tagName: "TABLE", className: "inner_tdc"});
+	  if( o ) xbinsert( o.parentNode, 1 );
+	  propagate=false;
+	  break;
+
+	case 188:		// ctrl+shift+,: insert new block above
+	  o=find_parent(src, {tagName: "TABLE", className: "inner_tdc"});
+	  if( o ) xbinsert( o.parentNode, -1 );
+	  propagate=false;
+	  break;
+
+	case 68:		// ctrl+shift+d: delete
+	  o=find_parent(src, {tagName: "TABLE", className: "inner_tdc"});
+	  if( o ) xdelete( o.parentNode, 1 );
+	  propagate=false;
+	  break;
+	}
+      }
+    }
+
+    if( propagate ) {
+      if( !e.stopPropagation ) {
+	e.cancelBubble = false;
+	e.returnValue = true;
+      }
+    } else {
+      //e.stopPropagation works in Firefox.
+      if (e.stopPropagation) {
+	e.stopPropagation();
+	e.preventDefault();
+      } else {
+	e.cancelBubble = true;
+	e.returnValue = false;
+      }
+    }
+
+    return propagate;
+  };
+
+  //Attach the function with the event
+  if(el.addEventListener)
+    el.addEventListener('keydown', func, false);
+  else if(el.attachEvent)
+    el.attachEvent('onkeydown', func);
+  else
+    el['onkeydown'] = func;
+
+  // now install the onchange handler
+  func = function(e) {
+    e = e || window.event;
+
+    var src=e.target!=null?e.target:e.srcElement;
+    if( src==null ) return false;
+    var c;
+    c=find_parent(src, {id: /^form(\d+)$/});
+    if( c!=null ) {
+      c=c.id.match(/^form(\d+)$/)[1];
+    } else {
+      return false;
+    }
+
+    //Find Which key is pressed
+    if (typeof(e.keyCode)=='number') code = e.keyCode;
+    else if (typeof(e.which)=='number') code = e.which;
+    else return xchanged(c);
+
+    if( code==0 ) return false;
+
+    if(!e.ctrlKey && !e.altKey && !e.metaKey || e.type=='change') {
+      if( code==Event.KEY_RETURN && e.type!='change' ) {
+	if( src.tagName=='INPUT' ) return false;
+      }
+
+      if( code==Event.KEY_TAB ||
+	  16<=code && code<=20 ||    // ctrl, shift, alt, caps, pause
+	  33<=code && code<=40 ||    // left, right, up, down, pageup, pagedown, insert, home, end
+	  code==45 ||		     // insert
+	  144<=code && code<=145 ||  // scroll, numlock
+	  code==91 || code==93 )     // menu, windows
+	return false;
+
+      return xchanged(c);
+    }
+    return false;
+  };
+
+  //Attach the function with the event
+  if(el.addEventListener) {
+    el.addEventListener('keyup', func, false);
+    el.addEventListener('change', func, false);
+  } else if(el.attachEvent) {
+    el.attachEvent('onkeyup', func);
+    el.attachEvent('onchange', func);
+  } else {
+    el['onkeyup'] = func;
+    el['onchange'] = func;
+  }
+}
+
+function add_inner_shortcuts( counter ) {
+}
+
 function add_resizer( counter ) {
-  var resizer=(typeof(counter)=='number' ? $('div'+counter) : $(counter))
-                .getElementsByTagName('div');
+  var resizer;
+
+  if( typeof(counter)=="string" || typeof(counter)=="number" ) {
+    resizer=$('div'+counter);
+  } else {
+    resizer=$(counter);
+  }
+  resizer=resizer.getElementsByTagName('table');
 
   for( var i=0; i<resizer.length; i++ ) {
-    if( resizer[i].className=='resizer' ) {
-      new Draggable(resizer[i],
-		    { constraint: 'vertical',
-		      resizing_element: get_prev_sibling(resizer[i], 'table'),
-		      starteffect: function(o) {
-			this.resizing_height=this.resizing_element.offsetHeight;
-		      },
-		      change: function(o) {
-			this.resizing_delta=o.currentDelta()[1];
-		      },
-		      endeffect: function(o) {
-			this.resizing_element.style.height=
-			  (this.resizing_height+this.resizing_delta)+'px';
-			o.style.top=0;
-		      },
-		      snap: function(x,y) {
-			if( this.resizing_height+y < 1 ) y=-this.resizing_height+1;
-			return [x,y];
-		      }
-		    });
+    if( resizer[i].className=='inner_tdc' ) {
+      new Resizeable( resizer[i], 
+		      {top: 0, left: 0, bottom: 8, right: 0} );
     }
   }
 }
 
+function set_focus_to_first_input(counter, where) {
+  var focus;
+
+  if( typeof(counter)=="string" || typeof(counter)=="number" ) {
+    focus=$('div'+counter);
+  } else {
+    focus=$(counter);
+  }
+
+  if( where==null ) {
+    focus=focus.getElementsByTagName('input');
+
+    for( var i=0; i<focus.length; i++ ) {
+      if( focus[i].type=='text' ) {
+	focus[i].focus();
+	return;
+      }
+    }
+    return;
+  } else {
+    var trs=focus.getElementsByTagName('tr');
+    var n=0;
+    for( var i=0; i<trs.length; i++ ) {
+      if( trs[i].className=='tdc' && n==where ) {
+	trs[i].getElementsByTagName('textarea')[0].focus();
+	return;
+      } else if( trs[i].className=='tdc' ) n++;
+    }
+    // not found so far. Set focus to the last textarea
+    trs[trs.length-1].getElementsByTagName('textarea')[0].focus();
+    return;
+  }
+}
+
+function focus2index(o) {
+  if( !o ) return null;
+  o=find_parent(o, {tagName: 'TR', className: 'tdc'});
+  if( !o ) return null;
+  var trs=o.parentNode.getElementsByTagName('TR');
+  var n=0;
+  for( var i=0; i<trs.length; i++ ) {
+    if( trs[i]==o ) return n;
+    if( trs[i].className=='tdc' ) n++;
+  }
+  return null;
+}
+
 function xopen( counter ) {
   var data=get_data(counter);
   if( $('div'+counter).innerHTML.length>0 ) {
@@ -58,8 +280,10 @@ function xopen( counter ) {
       //Element.show( 'reload'+counter );
     }
     Element.update( 'a'+counter, opener[2] );
+    set_focus_to_first_input( counter );
   } else {
     Element.update( 'a'+counter, opener[1] );
+    Element.show( 'progress' );
     new Ajax.Updater( { success: 'div'+counter },
 	  	      'index.html',
 		      { method: 'post',
@@ -79,6 +303,8 @@ function xopen( counter ) {
 			    $('reload'+counter).style.visibility='';
 			    //Element.show( 'reload'+counter );
 			    Element.update( 'a'+counter, opener[2] );
+			    add_inner_shortcuts( counter );
+			    set_focus_to_first_input( counter );
 			  } else {
 			    Element.update( 'a'+counter, opener[0] );
 			    var err;
@@ -97,15 +323,18 @@ function xopen( counter ) {
 				    req.status+")");
 			    }
 			  }
+			  Element.hide( 'progress' );
 			}
 		      } );
   }
 }
 
-function xreload( counter, o ) {
-  o.blur();
+function xreload( counter, o, focus ) {
+  if(o) o.blur();
+  focus=focus2index(focus);
   var data=get_data(counter);
   Element.update( 'a'+counter, opener[1] );
+  Element.show( 'progress' );
   new Ajax.Updater( { success: 'div'+counter },
 	  	      'index.html',
 		    { method: 'post',
@@ -127,6 +356,8 @@ function xreload( counter, o ) {
 		          Element.update( 'a'+counter, opener[2] );
 			  var f=$('form'+counter);
 			  update_header(counter, f.newkey.value, f.newuri.value);
+			  add_inner_shortcuts( counter );
+			  set_focus_to_first_input( counter, focus );
 		        } else {
 		          Element.update( 'a'+counter, opener[2] );
 			  var err;
@@ -145,6 +376,7 @@ function xreload( counter, o ) {
 				  req.status+")");
 			  }
 		        }
+			Element.hide( 'progress' );
 		      }
 		    } );
   return false;
@@ -155,10 +387,11 @@ function xclose( counter ) {
   $('reload'+counter).style.visibility='hidden';
   //Element.hide( 'reload'+counter );
   Element.update( 'a'+counter, opener[0] );
+  $('a'+counter).focus();
 }
 
 function xtoggle( counter, o ) {
-  o.blur();
+  if(o) o.blur();
   if( Element.visible( 'div'+counter ) ) {
     xclose( counter );
   } else {
@@ -192,9 +425,10 @@ function xreorder( counter ) {
 	oldblock=tds[i].getAttribute("ADM_BLOCK");
 	order=-1;
       }
+      var hi=tds[i].getElementsByTagName("input")[0];
       var ta=tds[i].getElementsByTagName("textarea");
       var tc;
-      if( ta.length>1 ) tc=ta[1];
+      if( ta && ta.length>1 ) tc=ta[1];
       ta=ta[0];
       var blk=tds[i].getAttribute("ADM_BLOCK");
       var ord=tds[i].getAttribute("ADM_ORDER");
@@ -211,6 +445,8 @@ function xreorder( counter ) {
       //    order+" id="+id+"\n");
       ta.name="action_"+oldblock+"_"+block+"_"+ord+"_"+order+"_"+id;
       if( tc ) tc.name="note_"+block+"_"+order;
+      hi.name="ysize_"+block+"_"+order;
+      hi.value=Element.getHeight(tds[i])-(Prototype.Browser.IE ? 4 : 0);
     }
   }
 }
@@ -224,8 +460,9 @@ function update_header( counter, key, uri ) {
   }
 }
 
-function xupdate( counter, o ) {
-  o.blur();
+function xupdate( counter, o, focus ) {
+  if(o) o.blur();
+  focus=focus2index(focus);
   xreorder( counter );
   var params=$('form'+counter).getElements().inject
     ({}, function(hash, element) {
@@ -234,14 +471,11 @@ function xupdate( counter, o ) {
        var method = element.tagName.toLowerCase();
        var parameter = Form.Element.Serializers[method](element);
 
-       if (parameter) {
-	 var key = encodeURIComponent(parameter[0]);
+       if (parameter && parameter.length) {
+	 var key = element.name;
 	 if (key.length == 0) return hash;
 
-	 if (parameter[1].constructor != Array)
-	   parameter[1] = [parameter[1]];
-
-	 hash[key]=parameter[1];
+	 hash[key]=parameter;
        }
        return hash;
      });
@@ -251,6 +485,7 @@ function xupdate( counter, o ) {
   params["key"]=d.key;
   params["uri"]=d.uri;
   Element.update( 'a'+counter, opener[1] );
+  Element.show( 'progress' );
   new Ajax.Updater( { success: 'div'+counter },
 		    'index.html',
                     { method: 'post',
@@ -264,6 +499,7 @@ function xupdate( counter, o ) {
 			  var f=$('form'+counter);
 			  set_data( counter, f.newkey.value, f.newuri.value );
 			  update_header(counter, f.newkey.value, f.newuri.value);
+			  add_inner_shortcuts( counter );
 			} else {
 			  var err;
 			  var errcode;
@@ -284,32 +520,29 @@ function xupdate( counter, o ) {
 				  req.status+")");
 			  }
 			}
-			Element.update( 'a'+counter, opener[2] );
+			var el=$('a'+counter);
+			if( el ) {
+			  Element.update( el, opener[2] );
+			  set_focus_to_first_input( counter, focus );
+			}
+			Element.hide( 'progress' );
 		      }
                     } );
   return false;
 }
 
-function find_parent_by_tag( o, tag ) {
-  tag=tag.toUpperCase();
-  while( o && o.nodeName != tag ) {
-    o=o.parentNode;
-  }
-  return o;
-}
-
 function get_form_counter( o ) {
-  return find_parent_by_tag( o, 'form' ).getAttribute("ADM_COUNTER");
+  return find_parent( o, {tagName: 'FORM'} ).getAttribute("ADM_COUNTER");
 }
 
 function xinsert( o, where ) {
-  o.blur();
-  var tr=find_parent_by_tag( o, 'tr' );
+  if(o) o.blur();
+  var tr=find_parent( o, {tagName: 'TR', className: 'tdc'} );
   var newnode=tr.cloneNode(true);
-  
+
   var ta=newnode.getElementsByTagName("textarea");
   ta[0].value='';
-  if( ta.length>1 ) ta[1].value='';
+  if( ta && ta.length>1 ) ta[1].value='';
 
   var hidden=newnode.getElementsByTagName("td")[0];
   hidden.setAttribute("ADM_ORDER", "");
@@ -329,11 +562,13 @@ function xinsert( o, where ) {
 
   $('save'+get_form_counter(tr)).style.visibility='';
 
+  newnode.getElementsByTagName('textarea')[0].focus();
+
   return false;
 }
 
 function find_next_free_form_block( o ) {
-  var form=find_parent_by_tag( o, 'form' );
+  var form=find_parent( o, {tagName: 'FORM'} );
   var rc=form.getAttribute("ADM_NBLOCKS");
   form.setAttribute("ADM_NBLOCKS", rc+1);
   return rc;
@@ -342,7 +577,7 @@ function find_next_free_form_block( o ) {
 function get_tr_block( o ) {
   //debug(o.inspect()+" class="+o.className+"\n");
   var rc=o.getElementsByTagName("td");
-  if( rc.length==0 ) return -1;
+  if( rc && rc.length==0 ) return -1;
   rc=rc[0].getAttribute("ADM_BLOCK");
   if( rc==null ) return -1;
   else return rc;
@@ -365,14 +600,14 @@ function get_prev_sibling( tr, what ) {
 }
 
 function update_bg( o ) {
-  var tbl=find_parent_by_tag( o, 'table' );
+  var tbl=find_parent( o, {tagName: 'TABLE'} );
   var trs=tbl.getElementsByTagName("tr");
   var style=-1;
   var block=-1;
 
   for (var i=0; i<trs.length; i++) {
     var td=trs[i].getElementsByTagName("td");
-    if( td.length && td[0].className.match(/^tdc\d+$/) ) {
+    if( td && td.length && td[0].className.match(/^tdc\d+$/) ) {
       if( get_tr_block(trs[i])!=block ) {
 	block=get_tr_block(trs[i]);
 	style=(style+1)%3;
@@ -393,13 +628,13 @@ function xbdelete( counter ) {
 }
 
 function xbinsert( o, where ) {
-  o.blur();
-  var tr=find_parent_by_tag( o, 'tr' );
+  if(o) o.blur();
+  var tr=find_parent( o, {tagName: 'TR', className: 'tdc'} );
   var newnode=tr.cloneNode(true);
   
   var ta=newnode.getElementsByTagName("textarea");
   ta[0].value='';
-  if( ta.length>1 ) ta[1].value='';
+  if( ta && ta.length>1 ) ta[1].value='';
 
   var hidden=newnode.getElementsByTagName("td")[0];
   hidden.setAttribute("ADM_BLOCK", find_next_free_form_block( tr ));
@@ -436,18 +671,44 @@ function xbinsert( o, where ) {
   $('save'+get_form_counter(tr)).style.visibility='';
   //Element.show( 'save'+get_form_counter(tr) );
 
+  newnode.getElementsByTagName('textarea')[0].focus();
+
   return false;
 }
 
 function xdelete( o ) {
-  o.blur();
-  var tr=find_parent_by_tag( o, 'tr' );
-  var form=find_parent_by_tag( tr, 'form' );
+  if(o) o.blur();
+  var tr=find_parent( o, {tagName: 'TR'} );
+  var form=find_parent( tr, {tagName: 'FORM'} );
   var parent=tr.parentNode;
 
+  var newfocus;
+  for( newfocus=tr.nextSibling; newfocus; newfocus=newfocus.nextSibling ) {
+    if( newfocus.tagName=='TR' ) break;
+  }
+  if( !newfocus ) {
+    for( newfocus=tr.previousSibling; newfocus; newfocus=newfocus.previousSibling ) {
+      if( newfocus.tagName=='TR' ) break;
+    }
+  }
+  if( newfocus ) {
+    var el=newfocus.getElementsByTagName("textarea");
+    if( el && el.length ) {
+      newfocus=el[0];
+    } else {
+      el=form.getElementsByTagName("input");
+      for( var i=0; i<el.length; i++ ) {
+	if( el[i].type=='text' ) {
+	  newfocus=el[i];
+	  break;
+	}
+      }
+    }
+  }
+
   var hidden=tr.getElementsByTagName("td")[0];
   var v=hidden.getAttribute("ADM_ID");
-  if( v.length ) {		// need to delete from database
+  if( v && v.length ) {		// need to delete from database
     var newnode=document.createElement('input');
     newnode.name=("delete_"+hidden.getAttribute("ADM_BLOCK")+"_"+
 		  hidden.getAttribute("ADM_ORDER")+"_"+v);
@@ -462,6 +723,7 @@ function xdelete( o ) {
 
   $('save'+get_form_counter(parent)).style.visibility='';
   //Element.show( 'save'+get_form_counter(parent) );
+  newfocus.focus();
 
   return false;
 }
@@ -481,14 +743,15 @@ function check_key( key ) {
 }
 
 function xnewkey( o, uri ) {
-  o.blur();
+  if(o) o.blur();
   if( next_counter!=null ) {
     var key='newkey';
     for( var i=1; check_key(key); i++ ) key='newkey'+i;
     if( uri==null ) uri="subroutine";
-    o=find_parent_by_tag(o, 'h2');
+    o=find_parent(o, {tagName: 'H2'});
 
     var newnode=document.createElement('div');
+    var newfocus=newnode;
     newnode.id='div'+next_counter;
     newnode.style.display='';
     newnode.setAttribute("ADM_KEY", key);
@@ -500,13 +763,11 @@ function xnewkey( o, uri ) {
 	'	  onsubmit="return false;"'+
 	'	  ADM_COUNTER="'+next_counter+'"'+
 	'	  ADM_NBLOCKS="1">'+
-	'	<table>'+
+	'	<table class="tdc">'+
 	'		<tr>'+
 	'			<td class="tdcol1">New Key:</td>'+
 	'			<td class="tdcol2">'+
-	'				<input type="text" name="newkey" id="key'+next_counter+'" value="'+key.escapeHTML()+'"'+
-	'					   onchange="return xchanged( '+next_counter+' );"'+
-	'					   onkeyup="return xchanged( '+next_counter+' );">'+
+	'				<input type="text" name="newkey" id="key'+next_counter+'" value="'+key.escapeHTML()+'">'+
 	'			</td>'+
 	'		</tr>'+
 	(uri==":PRE:"
@@ -514,34 +775,28 @@ function xnewkey( o, uri ) {
 	 : ('		<tr>'+
 	    '			<td class="tdcol1">New Uri:</td>'+
 	    '			<td class="tdcol2">'+
-	    '				<input type="text" name="newuri" id="uri'+next_counter+'" value="'+uri.escapeHTML()+'"'+
-	    '					   onchange="return xchanged( '+next_counter+' );"'+
-	    '					   onkeyup="return xchanged( '+next_counter+' );">'+
+	    '				<input type="text" name="newuri" id="uri'+next_counter+'" value="'+uri.escapeHTML()+'">'+
 	    '			</td>'+
 	    '		</tr>'))+
 	'		<tr><th colspan="2"><br>Action</th></tr>'+
 	'		<tr class="tdc">'+
 	'			<td ADM_BLOCK="0" ADM_ORDER="" ADM_ID=""'+
 	'				colspan="2" class="tdc1">'+
+	'			    <input type="hidden" name="ysize">'+
 	'		            <table class="inner_tdc">'+
 	'		                <tr>'+
-	'                                   <td>'+
+	'                                   <td class="fetch_ta">'+
 	'			                <textarea class="fetch_ta" rows="1" wrap="off"'+
-	'			                          title="action field"'+
-	'				                  onchange="return xchanged( '+next_counter+' );"'+
-	'						  onkeyup="return xchanged( '+next_counter+' );"></textarea>'+
+	'			                          title="action field"></textarea>'+
 	'			            </td>'+
 	(can_notes
-	 ? ('                               <td>'+
+	 ? ('                               <td class="comment_ta">'+
 	    '			                <textarea class="comment_ta" rows="1" wrap="word"'+
-	    '			                          title="comment field"'+
-	    '						  onchange="return xchanged( '+next_counter+' );"'+
-	    '						  onkeyup="return xchanged( '+next_counter+' );"></textarea>'+
+	    '			                          title="comment field"></textarea>'+
 	    '                               </td>')
 	 : '')+
 	'                               </tr>'+
 	'                           </table>'+
-	'			    <div class="resizer" name="resizer" title="resize text area above"></div>'+
 	'			</td>'+
 	'			<td class="control">'+
 	'				&nbsp;<a href="#" onclick="return xinsert(this, -1);"'+
@@ -595,9 +850,33 @@ function xnewkey( o, uri ) {
 	'<span class="header" id="header'+next_counter+'"></span>');
 
     o.parentNode.insertBefore( newnode, o.nextSibling );
+    //add_div_shortcuts(next_counter);
     update_header(next_counter, key, uri);
 
+    newfocus=newfocus.getElementsByTagName("input");
+    for( var i=0; i<newfocus.length; i++ ) {
+      if( newfocus[i].type=='text' ) {
+	newfocus[i].focus();
+	break;
+      }
+    }
+
     next_counter++;
   }
   return false;
 }
+
+function ie_height(o) {
+  if( Resizeable.current ) {
+    for( var i=o.parentNode; i; i=i.parentNode ) {
+      if( i==Resizeable.current.element ) {
+        return ((Resizeable.current.currentHeight>47
+                 ? Resizeable.current.currentHeight-7
+	         : 40)+"px");
+      }
+    }
+  }
+  return ((o.parentNode.offsetHeight>47
+	   ? o.parentNode.offsetHeight-7
+	   : 40)+"px");
+}
